// A concurrent prime sieve
package main

// remember typing :
// chan<- int: used only for sending
// <-chan int: used only for receiving

// Create and return a channel ch 
// where the sequence 2, 3, 5, 7, 9, ... will be sent by Gen
func Generate() <-chan int {
	out := make(chan int)
	go Gen(out)
	return out
}

// Send the sequence 2, 3, 5, 7, 9, ... on ch
func Gen(out chan<- int) {
	out <- 2
	// note the missing guard!
	for i := 3; ; i = i + 2 {
		out <- i
	}	
}

// receives the list of primes and prints each of them
func Rec(in <-chan int, count int) {
	for i := 1; i <= count; i++ {
		// receives a prime
		prime := <-in
		println(prime,"is prime n.",i)
		// install a new filter for the newly received prime
		in = Filter(in, prime) // updates in
	}
}

// Copy the values from channel 'in' to channel 'out'
// removing those divisible by 'prime'
func Filter(in <-chan int, prime int) <-chan int {
	out := make(chan int)
	go Fil(in,out,prime)
	return out
}

func Fil(in <-chan int, out chan<- int, prime int) {
	println("--- Created filter", prime, "from", in, "to", out)
	for {
		i := <-in // Receive value from 'in'
		println("--- Filter", prime, "receives", i, "on", in)
		if i%prime != 0 {
			println("--- Filter", prime, "sends", i, "to", out)
			out <- i // Send 'i' to 'out'
		} else {
			println("--- Filter", prime, "discards", i)
		}
	}
}

// The prime sieve: Daisy-chain Filter processes.
func main() {
	ch := Generate() // Launch Generate goroutine.
	Rec(ch,40)
}
